/////////////////////////////////////////////////////////////////////////////////////
// ADSCopy.cpp
// 710020-1112x
//
// 5/20/02
// Applied Data Systems
//
// ADSCopy aids in setting the initial state of your CE device when coming back from 
// a hard-reset or power up.  It is used to automatically copy and/or register files 
// at boot time.
//
// Usage:
// 1. Place ADSCopy.exe in the \FlashFX Disk\Startup directory.
// 2. Place a file ADSCopy.inf in the \FlashFX Disk\ADSCopy directory.
// 3. Place files to be copied/registered in the \FlashFX Disk\ADSCopy directory.
// 
//    On startup, ADSCopy.exe will parse ADSCopy.inf.
//  
//    - ADSCopy must be a UNICODE text file.
//    - Each file name must be on it's own line, with a new line (\n) terminator (including the last line!)
//    - The filename can be suffixed with -r to have the file registered (DllRegisterServer called)
//    - The filename can be suffiexed with -e to have the file executed
//    - Filenames and paths must be enclosed in single quotes
//
// Example:
// To copy and register MyLibrary.DLL to the Windows directory, then run copy and run MyApp.exe
// the following would reside in ADSCopy.inf:
//
//    'MyLibrary.dll' '\Windows' -r
//    'MyApp.exe' '\Program Files\MyApp' -e 
//    
// History:
// ---------
// 7/9/02 - ctacke
//   Modified to allow command for waiting for executed process to finish.  The execution is now somewhat
//   altered.  Now, any line with a flag will *only* act upon the flag - file names with a flag will *not*
//   be automatically copied.  To copy and then run, you must use two separate lines in the INF file.
//
//  New Flag Definitions - **any line with a flag will not get copied** (see above).
//    -r : Register.  Remains the same, this will call DllRegisterServer on the file
//    -e : Execute and continue.  Simply executes the named file and continues processing
//    -x : Execute and wait.  Executes the named file a waits infinitely for the spawned process
//         to complete before continuing.  
//  
// ---------
// 7/19/02 - ctacke
//   Modified to accept running files with command-line parameters.
//   Modified to no longer run always from FlashFX Disk.  Instead the source directory for files
//   to be copied is now {exe location}\ADSCopy.  THis allows ADSCopy to be used effectively 
//   from FlashFX or ATA card.
//   Note that this will cause previous file location setups to fail.  For FlashFX disk install 
//   described above, the source directory would now be \FlashFX Disk\Startup\ADSCopy.
//
// ---------
// 7/22/02 - ctacke
//   Modified to allow a space in the file path *and* use command-line parameters in the same inf line.
//   If a file's path or name contains a space, the entire path must be enclosed in double-quotes.
//   For example, to run ADSAlloc from the FlashFX Disk directory with an 8MB allocation, use the following
//   INF file line:
//
//			'"\FlashFX Disk\ADSAlloc.exe" 0x800' -e
// 
// ---------
// 7/26/02 - ctacke
//   Previous versions failed if the target directory existed.  Added code to create target directories.
//   Removed requirement for Unicode INF file (Unicode will now cause an error) and tightened up parsing routines
//   code is now more memory and processor efficient.
// 
// ---------
// 8/14/02 - ctacke
//   Added code to ensure only 1 instance may run at a time.  Also added code to make sure it only runs
//   on first boot, not after a restart.
//
// ---------
// 12/2/02 - ctacke
//   Added QUIT_ON_COPY_FAILURE flag to allow compiling so the app doesn't end if a file copy fails (allows
//   filenames to be added to the INF that may not exist in the source directory)
//   Modified ParseLine so that the file path is inf extracted, not assumed
//
//
// ---------
// 06/11/03 710020-11127 - ctacke
//   Moved copy operations to a separate thread and added a status window to prevent apparent device startup problems
//   Created CopyFileB which blocks and is more reliable than the CopyFile API
//
// ---------
// July 18, 03 710020-11128 - ctacke
//   - Added adscopy.log file support to determine hard/soft boot on devices
//     with persistent registries
//   - added support for INF comment lines (lines starting with ";")
//   - added code to trim leading spaces in INF lines
//   - fixed bug that caused crash with empty INF lines
//   - added support for full directory copy wildcards
//
// ---------
// Sept 11, 03 710020-11129 - ctacke
//   - Bug fix: added calls to CoInitializeEx, CoUninitialize in RegisterFile function
//
// ---------
// July 23, 04 - ctacke
//   - modified source file parsing.  Files starting with '\' are absolute, otherwise
//     assumed to be relative
//
// ---------
// Sept 9, 04 - ctacke
//   - Added check for directories to prevent error when trying to copy them
//   - major cleanup up of line parsing - better wildcard, relative path support
//
// Sept 13, 04 - ctacke
//   - Custom modifications for Communitek
//
/////////////////////////////////////////////////////////////////////////////////////
#include <windows.h>
#include "winreg.h"
#include "resource.h"
#include "objbase.h"

// run switches
#define QUIT_ON_COPY_FAILURE   FALSE

// function prototypes
BOOL ExecuteFile(TCHAR *FileName, BOOL bWait);
BOOL IsCabFile(TCHAR *FileName);
BOOL IsVBFile(TCHAR *FileName);
BOOL ParseLine(TCHAR *linedata, TCHAR *filename, TCHAR *destination, TCHAR *flag);
BOOL RegisterFile(TCHAR *FileName);
BOOL ExecuteFile(TCHAR *FileName);
void GetCurrentDirectory(TCHAR *szDirectory);
void VerifyDirectory(TCHAR *filename);
DWORD WINAPI CopyThreadProc(LPVOID lpParameter);
BOOL CALLBACK MainWndProc(HWND hWnd, UINT wMsg, WPARAM wParam, LPARAM lParam);
BOOL CopyFileB(LPCTSTR lpExistingFileName, LPCTSTR lpNewFileName);
void DirectoryName(TCHAR *filename, TCHAR *directoryname);

// global variables
TCHAR	CopyPath[MAX_PATH];
TCHAR	INFFile[MAX_PATH];
HANDLE	hMutex;

/************************* CopyThreadProc Method *************************/
DWORD WINAPI CopyThreadProc(LPVOID lpParameter)
{
	FILE *stream;
	DWORD dwStartPos;

	// get the running directory
	GetCurrentDirectory(CopyPath);

	// build paths
	_tcscat(CopyPath, _T("\\ADSCopy"));
	_tcscpy(INFFile, CopyPath);
	_tcscat(INFFile, _T("\\ADSCopy.inf"));

	RETAILMSG(TRUE, (_T("Copying files from %s\r\n"), CopyPath));

	// Open INF file
	if((stream = _wfopen(INFFile, _T("rt"))) == NULL)
    {
		RETAILMSG(TRUE, (_T("ADSCopy: Can't find INF file: %s\r\nGetLastError returned: %d\r\n"), INFFile, GetLastError()));
		RETAILMSG(TRUE, (_T("----- ADSCopy Finished ----- \r\n\n")));

        return 3;
    }

	// parse file data
	dwStartPos = 1; //Set to 1 to skip unicode header

	TCHAR filename[MAX_PATH];
	TCHAR destination[MAX_PATH];
	TCHAR flags[MAX_PATH];
	TCHAR line[MAX_PATH];

	LoadCursor(NULL, IDC_WAIT);

	while(!feof(stream))
	{
		// clear buffer
		_tcscpy(line, _T(""));

		// checks for EOF
		if(_fgetts(line, MAX_PATH, stream) == NULL)
			continue;

		//sanity check
		if(ferror(stream))
		{
			RETAILMSG(TRUE, (_T("ADSCopy: Error Reading INF file: %s\r\nGetLastError returned: %d\r\n"), INFFile, GetLastError()));
			fclose(stream);
			RETAILMSG(TRUE, (_T("----- ADSCopy Finished ----- \r\n\n")));
			return 4;
		}

		DEBUGMSG(TRUE, (_T("Line: %s\r\n"), line));

		// strip leading spaces
		while(line[0] == ' ')
			_tcscpy(line, &line[1]);

		if(_tcslen(line) == 0)
			continue;
		
		switch(line[0])
		{
			case ';':	// lines starting with a ";" are comments
			case 13:	// ignore empty lines (just a \r)
			case 10:	// ignore empty lines (just a \n)
				continue;

		}

		if(!ParseLine(line, filename, destination, flags))
		{
			RETAILMSG(TRUE, (_T("----- ADSCopy Finished ----- \r\n\n"), INFFile, GetLastError()));
			return 5;
		}
		
		DEBUGMSG(TRUE, (_T("Parsed filename: %s\r\n"), filename));
		DEBUGMSG(TRUE, (_T("Parsed destination: %s\r\n"), destination));
		DEBUGMSG(TRUE, (_T("Parsed flags: %s\r\n"), flags));
		
		if(_tcsstr(flags, _T("-r")))
		{   // register
			DEBUGMSG(TRUE, (_T("Registering %s\r\n"), filename));

			RegisterFile(filename);
		}
		else if(_tcsstr(flags, _T("-e")))
		{   // run and return
			DEBUGMSG(TRUE, (_T("Execute %s and return\r\n"), filename));

			ExecuteFile(filename, false);
		}
		else if(_tcsstr(flags, _T("-x")))
		{   // run and wait
			DEBUGMSG(TRUE, (_T("Execute %s and wait\r\n"), filename));

			ExecuteFile(filename, true);
		}
		else
		{

			RETAILMSG(TRUE, (_T("ADSCopy: Copying from '%s'\r\nADSCopy:          to '%s'\r\n"), filename, destination));
			
			// ensure target directory exists
			VerifyDirectory(destination);

			WIN32_FIND_DATA fd;

			DEBUGMSG(TRUE, (_T("ADSCopy: Calling FindFirstFile(%s)\r\n"), filename));

			HANDLE hFind = FindFirstFile(filename, &fd);

			if(hFind != INVALID_HANDLE_VALUE)
			{
				do
				{
					DEBUGMSG(TRUE, (_T("ADSCopy: found %s\r\n"), fd.cFileName));

					for(int i = _tcslen(filename) ; i > 0 ; i--)
					{
						if(filename[i] == '\\')
						{
							filename[i] = '\0';
							break;
						}
					}

					_tcscat(filename, _T("\\"));
					_tcscat(filename, fd.cFileName);


					for(i = _tcslen(destination) ; i > 0 ; i--)
					{
						if(destination[i] == '\\')
						{
							destination[i] = '\0';
							break;
						}
					}

					_tcscat(destination, _T("\\"));
					_tcscat(destination, fd.cFileName);


					DWORD attributes = GetFileAttributes(filename);

					if(attributes == 0xFFFFFFFF)
					{
						RETAILMSG(TRUE, (_T("ADSCopy: cannot get attributes for %s\r\nGetLastError returned: %d\r\n"), filename, GetLastError()));
						continue;
					}

					if(attributes & FILE_ATTRIBUTE_DIRECTORY)
					{
						// *** we could recurse here to deep copy - future feature addition ***

						// only copy files, not dirs
						RETAILMSG(TRUE, (_T("ADSCopy: ignoring %s - it is a folder\r\n"), filename));
						continue;
					}


					DEBUGMSG(TRUE, (_T("ADSCopy: CopyFileB(\"%s\", \"%s\"\r\n"), filename, destination));
					if(!CopyFileB(filename, destination))
					{   // copy file
						RETAILMSG(TRUE, (_T("ADSCopy: Failed to copy file: %s\r\nGetLastError returned: %d\r\n"), filename, GetLastError()));
						if(QUIT_ON_COPY_FAILURE)
						{
							fclose(stream);
							RETAILMSG(TRUE, (_T("----- ADSCopy Finished ----- \r\n\n")));
							return 6;
						} // if(QUIT_ON_COPY_FAILURE)
					} // if(!CopyFile(filename, destination, FALSE))
				} while(FindNextFile(hFind, &fd));
			} // if(hFind != INVALID_HANDLE_VALUE)
			else
			{
				DEBUGMSG(TRUE, (_T("No match found\r\n")));
			}

			if(hFind != INVALID_HANDLE_VALUE)
				FindClose(hFind);

		} // else
	} // while(!feof(stream))
	
	LoadCursor(NULL, 0);

	return 0;
}

BOOL CopyFileB(LPCTSTR lpExistingFileName, LPCTSTR lpNewFileName)
{
	const DWORD buffersize = 0x2000;
	DWORD read		= 0;;
	DWORD written	= 0;

	HANDLE hSource = CreateFile(lpExistingFileName, 
								GENERIC_READ, 
								0, 
								NULL, 
								OPEN_EXISTING, 
								FILE_ATTRIBUTE_NORMAL, 
								NULL);

	if(hSource == INVALID_HANDLE_VALUE)
	{
		RETAILMSG(TRUE, (_T("ADSCopy: cannot open source file: %s\r\nGetLastError returned: %d\r\n"), lpExistingFileName, GetLastError()));
		return FALSE;
	}

	HANDLE hDest = CreateFile(lpNewFileName, 
								GENERIC_WRITE, 
								0, 
								NULL, 
								CREATE_ALWAYS, 
								FILE_ATTRIBUTE_NORMAL, 
								NULL);

	if(hDest == INVALID_HANDLE_VALUE)
	{
		RETAILMSG(TRUE, (_T("ADSCopy: cannot create destination file: %s\r\nGetLastError returned: %d\r\n"), lpExistingFileName, GetLastError()));
		return FALSE;
	}

	BYTE buffer[buffersize];

	BOOL r = TRUE;
	BOOL eof = FALSE;

	do
	{
		read = 0;
		written = 0;

		do
		{
			r = ReadFile(hSource, buffer, buffersize, &read, NULL);
		} while(!r);
		
		eof = (r && (read == 0));

		if(!eof)
		{
			do
			{
				r = WriteFile(hDest, buffer, read, &written, NULL);
			} while(!r);
		}
	} while(!eof);

	CloseHandle(hSource);
	CloseHandle(hDest);
	return TRUE;
}

BOOL CALLBACK MainWndProc(HWND hWnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	return FALSE;
}

/************************* WinMain Method *************************/
int WINAPI WinMain(
    HINSTANCE   hInstance,
    HINSTANCE   hInstPrev,
    LPTSTR      lpszCmdLine,
    int         nCmdShow)
{
	TCHAR *buffer = NULL;
	BOOL bLastFile = FALSE;

	RETAILMSG(TRUE, (_T("\r\n----- ADSCopy Starting -----\r\n")));

	// make the app a singleton
	hMutex = CreateMutex(NULL, true, _T("ADSCopy"));

	if(GetLastError() == ERROR_ALREADY_EXISTS)
	{
		RETAILMSG(TRUE, (_T("Another instance is running!\r\n")));
		RETAILMSG(TRUE, (_T("----- ADSCopy Finished ----- \r\n\n")));
		return 0;
	}

	// *** BEGIN CUSTOMER SPECIFIC MODIFICATIONS ***
	// If there is an existing rtablet.exe or ntablet.exe in the
	// \Flashdisk directory and the external storage card is present when 
	// the device is turned on then the existing exe(s) have to be deleted
	// before the exe(s) on the storage card are copied over.
	DeleteFile(_T("\\FlashFX Disk\\rtablet.exe"));
	DeleteFile(_T("\\FlashFX Disk\\ntablet.exe"));
	// *** END CUSTOMER SPECIFIC MODIFICATIONS ***

	// early versions of ADSCopy checked the HKLM\Comm\BootCount registry setting
	// since this key gets persisted on devices with a persistent registry we 
	// now create a file in RAM on first run that tells us whether we're coming
	// up from a hard rest or not

	DWORD dw	= 0;;
	DWORD data	= 0;

	FILE *log = _tfopen(_T("\\Temp\\adscopy.log"), _T("r+b"));

	int gle = GetLastError();

	if(log != 0)
	{	
		fread(&data, sizeof(DWORD), 1, log);
		fseek(log, 0, SEEK_SET);
		data++;
		fwrite(&data, sizeof(DWORD), 1, log);

		RETAILMSG(TRUE, (_T("This is execution %i\r\n"), data));
		RETAILMSG(TRUE, (_T("----- ADSCopy Finished ----- \r\n\n")));

		fclose(log);
		return 0;
	}

	data = 1;
	log = _tfopen(_T("\\Temp\\adscopy.log"), _T("w+b"));
	fwrite(&data, sizeof(DWORD), 1, log);
	fclose(log);

	RETAILMSG(TRUE, (_T("This is the first execution\r\n")));

	// create the dialog
	HWND hWndMain = CreateDialog(hInstance, MAKEINTRESOURCE(IDD_DIALOG), NULL, MainWndProc);

	HANDLE hThread = CreateThread(NULL, 0, CopyThreadProc, NULL, 0, 0);

	int progress = 0;
	TCHAR szProgress[32];
	_tcscpy(szProgress, _T(""));

	do
	{
		if(progress++ > 30)
		{
			progress = 0;
			_tcscpy(szProgress, _T(""));
		}
		_tcscat(szProgress, _T("-"));

		SetWindowText(GetDlgItem(hWndMain, IDC_PROGRESS), szProgress);
		
		UpdateWindow(hWndMain);
	} while(WaitForSingleObject(hThread, 500) == WAIT_TIMEOUT);

	// destroy the dialog
	DestroyWindow(hWndMain);

	RETAILMSG(TRUE, (_T("----- ADSCopy Finished ----- \r\n\n") ));

    return 0;
}

/************************* VerifyDirectory Method *************************/
void VerifyDirectory(TCHAR *filename)
{
	TCHAR dirname[MAX_PATH];
	TCHAR *pstart = filename;
	TCHAR *pend = _tcsstr(filename, _T("\\"));
	DWORD attr = 0;

	if(_tcslen(filename) <= 1) return;

	if(!(pend - pstart > 0))
	{ // doesn't start with a backslash
		_tcscpy(dirname, _T("\\"));
		pstart++;
	}

	while(true)
	{
		pend = _tcsstr(pstart, _T("\\"));

		if((unsigned)(pend - pstart) > _tcslen(pstart))
		{  // last directory
			_tcscat(dirname, filename);
			if(! _tcsstr(dirname, _T(".")))
			{
				attr = GetFileAttributes(dirname);
				if((attr == 0xffffffff) && ((GetLastError() == ERROR_PATH_NOT_FOUND) || (GetLastError() == ERROR_FILE_NOT_FOUND)))
				{
					DEBUGMSG(TRUE, (_T("Creating directory: %s\r\n"), dirname));
					CreateDirectory(dirname, NULL);
				}
			}
			break;
		}
		else
		{
			_tcsncpy(dirname, filename, (pend - filename));
			dirname[pend - filename] = '\0';
			attr = GetFileAttributes(dirname);
			if((attr == 0xffffffff) && ((GetLastError() == ERROR_PATH_NOT_FOUND) || (GetLastError() == ERROR_FILE_NOT_FOUND)))
			{
				DEBUGMSG(TRUE, (_T("Creating directory: %s\r\n"), dirname));
				CreateDirectory(dirname, NULL);
			}
			_tcscpy(dirname, _T("\\"));
			pstart = pend;
			pstart++;
		}
	}
}

/************************* ParseLine Method *************************/
BOOL ParseLine(TCHAR *linedata, TCHAR *filename, TCHAR *destination, TCHAR *flag)
{
	TCHAR *pStart;
	TCHAR *pEnd;
	TCHAR sourcename[MAX_PATH];

	DEBUGMSG(TRUE, (_T("---- Parsing line: %s\r\n"), linedata));

	// extract source file name
	pStart = _tcsstr(linedata, _T("'"));
	if(pStart == NULL)
	{
		RETAILMSG(TRUE, (_T("ADSCopy: Invalid INF line (no source starting quote):\r\n%s\r\n"), linedata));
		return FALSE;
	}

	pEnd = _tcsstr(pStart + 1, _T("'"));
	if(pEnd == NULL)
	{
		RETAILMSG(TRUE, (_T("ADSCopy: Invalid INF line (no source ending quote):\r\n%s\r\n"), linedata));
		return FALSE;
	}
	
	_tcscpy(filename, _T(""));
	_tcsncat(filename, pStart + 1, pEnd - pStart - 1);
	filename[pEnd - pStart] = '\0';

	DEBUGMSG(TRUE, (_T("   filename: %s\r\n"), filename));

	// if the filename starts with a backslash, the path is absolute
	// if not, it's relative to the ADSCopy subfolder
	if(filename[0] == '\\')
	{
		_tcscpy(sourcename, filename);
	}
	else
	{
		_tcscpy(sourcename, CopyPath);

		if(CopyPath[_tcslen(CopyPath) - 1] != '\\')
			_tcscat(sourcename, _T("\\"));

		_tcscat(sourcename, filename);
	}

	
	DEBUGMSG(TRUE, (_T("   sourcename: %s\r\n"), sourcename));

	// extract destination file path
	pStart = _tcsstr(pEnd + 1, _T("'"));

	// if no destination exists, default to the root
	if(pStart == NULL)
	{
		DEBUGMSG(TRUE, (_T("   no destination found\r\n")));
		_tcscpy(destination, _T("\\"));
		_tcscpy(flag, _T(""));
	}
	else
	{
		pEnd = _tcsstr(pStart + 1, _T("'"));
		
		if(pEnd == NULL)
		{
			RETAILMSG(TRUE, (_T("ADSCopy: Invalid INF line (no destination ending quote):\r\n%s\r\n"), linedata));
			return FALSE;
		}

		_tcscpy(destination, _T(""));
		_tcsncat(destination, pStart + 1, pEnd - pStart - 1);
		destination[pEnd - pStart - 1] = '\0';

		DEBUGMSG(TRUE, (_T("   destination: %s\r\n"), destination));

		if(filename[0] != '\\')
		{  // Don't copy over filename and paths, use from inf file, no assuming			
			// only add a terminating \ if it's missing
			if(destination[pEnd - pStart - 2] != 0x5C)
				_tcscat(destination, _T("\\"));

			_tcscpy(filename, sourcename);
		}

		DEBUGMSG(TRUE, (_T("   Parsed filename: %s\r\n"), filename));
	}
	
	// extract flags
	pStart = _tcsstr(pEnd + 1, _T("-"));
	
	if(pStart == NULL)
	{
		DEBUGMSG(TRUE, (_T("   no flags found\r\n")));

		_tcscpy(flag, _T(""));
		
		DEBUGMSG(TRUE, (_T("---- Done parsing\r\n")));

		return TRUE;
	}

	pEnd = &linedata[_tcslen(linedata)];
	_tcsncpy(flag, pStart, pEnd - pStart);
	flag[pEnd - pStart] = '\0';
	
	DEBUGMSG(TRUE, (_T("   flags: %s\r\n"), flag));
	DEBUGMSG(TRUE, (_T("---- Done parsing\r\n")));

	return TRUE;
}

/************************* RegisterFile Method *************************/
BOOL RegisterFile(TCHAR *FileName)
{
	HINSTANCE hLib = LoadLibrary(FileName);
	typedef LONG (*DLLREGISTERSERVER)(void);
	DLLREGISTERSERVER DllRegisterServer;

	if (hLib == NULL)
	{
		RETAILMSG(TRUE, (_T("ADSCopy: Failed to register file: %s\r\nGetLastError returned: %d\r\n"), FileName, GetLastError()));
		return FALSE;
	}

	DllRegisterServer = (DLLREGISTERSERVER) GetProcAddress(hLib,(L"DllRegisterServer"));

	if(DllRegisterServer == NULL)
	{
		RETAILMSG(TRUE, (_T("ADSCopy: Failed to register file: %s\r\nGetLastError returned: %d\r\n"), FileName, GetLastError()));
		FreeLibrary(hLib);
		return FALSE;
	}

	HRESULT hr = DllRegisterServer();
	if(hr != S_OK)
	{
		RETAILMSG(TRUE, (_T("ADSCopy: Failed to register file: %s\r\nDllRegisterServer returned: %d\r\n"), FileName, hr));
		FreeLibrary(hLib);
		return FALSE;
	}

	RETAILMSG(TRUE, (_T("ADSCopy: Successfully registered %s\r\n"), FileName));

	FreeLibrary(hLib);

	return TRUE;
}

/************************* RunFile Method *************************/
BOOL ExecuteFile(TCHAR *FileName, BOOL waitForCompletion)
{
	PROCESS_INFORMATION pi;
	
	if(IsCabFile(FileName))
	{
		RETAILMSG(TRUE, (_T("ADSCopy: installing CAB '%s'\r\n"), FileName));

		if(!CreateProcess(_T("\\Windows\\wceload.exe"), FileName, NULL, 0, FALSE, 0, 0, 0, NULL, &pi))
		{
			RETAILMSG(TRUE, (_T("ADSCopy: install failed: %s\r\nGetLastError returned: %d\r\n"), FileName, GetLastError()));
			return FALSE;
		}
	}
	else if(IsVBFile(FileName))
	{
		RETAILMSG(TRUE, (_T("ADSCopy: starting VB app '%s'\r\n"), FileName));

		if(!CreateProcess(_T("\\Windows\\pvbload.exe"), FileName, NULL, 0, FALSE, 0, 0, 0, NULL, &pi))
		{
			RETAILMSG(TRUE, (_T("ADSCopy: failed to run: %s\r\nGetLastError returned: %d\r\n"), FileName, GetLastError()));
			return FALSE;
		}
	}
	else
	{
		TCHAR *p;

		// check for command line params
		p = _tcsstr(FileName, _T(" ")); 
		if(p == NULL)
		{
			RETAILMSG(TRUE, (_T("ADSCopy: starting app '%s'\r\n"), FileName));	

			if(!CreateProcess(FileName, NULL, NULL, 0, FALSE, 0, 0, 0, NULL, &pi))
			{
				RETAILMSG(TRUE, (_T("ADSCopy: failed to run: %s\r\nGetLastError returned: %d\r\n"), FileName, GetLastError()));
				return FALSE;
			}
		}
		else
		{  // a space exists in the path.  
			TCHAR *q = NULL;
			TCHAR params[MAX_PATH];

			// determine if there are quotes, which delimit the path
			q = _tcsstr(FileName, _T("\"")); 
			
			if(q == NULL)
			{   // no quotes
				//the space separates the path and cmd line params
				// copy the params
				_tcscpy(params, p + 1);

				// crop from filename
				*p = '\0';
			}
			else
			{	// section between quotes is the path

				q++;
				q = _tcsstr(q, _T("\"")); 

				if(q == NULL)
				{
					RETAILMSG(TRUE, (_T("ADSCopy: Invalid INF line: missing double quote\r\n")));
					return FALSE;
				}

				q+=2;

				// copy the params
				_tcscpy(params, q);
			
				// crop the filename, strip quotes
				q-=2;
				*q = '\0';
				q = FileName;
				q++;
				_tcscpy(FileName, q);
			}

			RETAILMSG(TRUE, (_T("ADSCopy: starting app '%s' with command line '%s'\r\n"), FileName, params));	

			if(!CreateProcess(FileName, params, NULL, 0, FALSE, 0, 0, 0, NULL, &pi))
			{
				RETAILMSG(TRUE, (_T("ADSCopy: failed to run: %s\r\nGetLastError returned: %d\r\n"), FileName, GetLastError()));
				return FALSE;
			}
		}
	}

	if(waitForCompletion)
	{
		WaitForSingleObject(pi.hProcess, INFINITE);
	}

	return TRUE;
}

/************************* IsCabFile Method *************************/
BOOL IsCabFile(TCHAR *FileName)
{
	LPCTSTR ptrPos;
	// Check for existence of '.cab'
	ptrPos = _tcsstr(FileName, _T(".cab"));

	if (ptrPos == NULL)
	{
		// lowercase not found, check for upper case
		ptrPos = _tcsstr(FileName, _T(".CAB"));
	}

	return !(ptrPos == NULL);
}

/************************* IsVBFile Method *************************/
BOOL IsVBFile(TCHAR *FileName)
{
	LPCTSTR ptrPos;

	// Check for existence of '.vb'
	ptrPos = _tcsstr(FileName, _T(".vb"));

	if (ptrPos == NULL)
	{
		// lowercase not found, check for upper case
		ptrPos = _tcsstr(FileName, _T(".VB"));
	}

	return !(ptrPos == NULL);
}

/************************* GetCurrentDirectory Method *************************/
void GetCurrentDirectory(TCHAR *szDirectory)
{
	TCHAR *p, *p2;

	// get current directory
	GetModuleFileName(NULL, szDirectory, MAX_PATH);

	// trim off the exe name
	p = _tcsstr(szDirectory, _T("\\"));

	if(p == NULL)
	{
		// no slash found - return root?
		_tcscpy(szDirectory, _T("\\"));
		
		return;
	}

	while( p != NULL)
	{
		p2 = p + 1;
		p = _tcsstr(p2, _T("\\"));
	}

	// null terminate to crop
	p2--;
	*p2 = '\0';

	DEBUGMSG(TRUE, (_T("GetCurrentDirectory = %s\r\n"), szDirectory));

	return;
}
